// Copyright 2018 The Cockroach Authors.
//
// Use of this software is governed by the Business Source License
// included in the file licenses/BSL.txt.
//
// As of the Change Date specified in that file, in accordance with
// the Business Source License, use of this software will be governed
// by the Apache License, Version 2.0, included in the file
// licenses/APL.txt.

package main

import (
	"bytes"
	"flag"
	"fmt"
	"io"
	"io/ioutil"
	"os"
	"path/filepath"
	"regexp"

	"github.com/cockroachdb/cockroach/pkg/sql/colexec/execgen"
	"github.com/cockroachdb/cockroach/pkg/sql/colexecbase/colexecerror"
	"github.com/cockroachdb/errors"
	"github.com/cockroachdb/gostdlib/x/tools/imports"
)

var (
	errInvalidArgCount = errors.New("invalid number of arguments")
)

func main() {
	gen := execgenTool{stdErr: os.Stderr}
	if !gen.run(os.Args[1:]...) {
		os.Exit(2)
	}
}

type execgenTool struct {
	// fmtSources runs the go fmt tool on code generated by execgenTool, if this
	// setting is true.
	fmtSources bool

	// stdErr is the writer to which all standard error output will be redirected.
	stdErr io.Writer

	// cmdLine stores the set of flags used to invoke the Execgen tool.
	cmdLine *flag.FlagSet
	verbose bool
}

// generator is a func that, given an input file's contents as a string,
// outputs the result of execgen to the outputFile.
type generator func(inputFileContents string, outputFile io.Writer) error

var generators = make(map[string]entry)

type entry struct {
	fn        generator
	inputFile string
}

func registerGenerator(g generator, outputFile, inputFile string) {
	if _, ok := generators[outputFile]; ok {
		colexecerror.InternalError(errors.AssertionFailedf("%s generator already registered", outputFile))
	}
	generators[outputFile] = entry{fn: g, inputFile: inputFile}
}

func (g *execgenTool) run(args ...string) bool {
	// Parse command line.
	var printDeps bool
	g.cmdLine = flag.NewFlagSet("execgen", flag.ContinueOnError)
	g.cmdLine.SetOutput(g.stdErr)
	g.cmdLine.Usage = g.usage
	g.cmdLine.BoolVar(&g.fmtSources, "fmt", true, "format and imports-process generated code")
	g.cmdLine.BoolVar(&g.verbose, "verbose", false, "print out debug information to stderr")
	g.cmdLine.BoolVar(&printDeps, "M", false, "print the dependency list")
	err := g.cmdLine.Parse(args)
	if err != nil {
		return false
	}

	// Get remaining args after any flags have been parsed.
	args = g.cmdLine.Args()
	if len(args) < 1 {
		g.cmdLine.Usage()
		g.reportError(errInvalidArgCount)
		return false
	}

	for _, out := range args {
		_, file := filepath.Split(out)
		e := generators[file]
		if e.fn == nil {
			g.reportError(errors.Errorf("unrecognized filename: %s", file))
			return false
		}
		if printDeps {
			if e.inputFile == "" {
				// This output file has no template dependency (its template
				// is embedded entirely in execgenTool). Skip it.
				continue
			}
			fmt.Printf("%s: %s\n", out, e.inputFile)
		} else {
			if err := g.generate(out, e); err != nil {
				g.reportError(err)
				return false
			}
		}
	}
	return true
}

var emptyCommentRegex = regexp.MustCompile(`[ \t]*//[ \t]*\n`)
var emptyBlockCommentRegex = regexp.MustCompile(`[ \t]*/\*[ \t]*\*/[ \t]*\n`)

func (g *execgenTool) generate(path string, entry entry) error {
	var buf bytes.Buffer
	buf.WriteString("// Code generated by execgen; DO NOT EDIT.\n")

	var inputFileContents string
	var err error
	if entry.inputFile != "" {
		inputFileBytes, err := ioutil.ReadFile(entry.inputFile)
		if err != nil {
			return err
		}
		inputFileContents, err = execgen.Generate(string(inputFileBytes))
		if err != nil {
			return err
		}
		if g.verbose {
			fmt.Fprintln(os.Stderr, "generated code before text/template runs")
			fmt.Fprintln(os.Stderr, "-----------------------------------")
			fmt.Fprintln(os.Stderr, inputFileContents)
			fmt.Fprintln(os.Stderr, "-----------------------------------")
		}
	}

	err = entry.fn(inputFileContents, &buf)
	if err != nil {
		return err
	}

	b := buf.Bytes()
	// Delete empty comments ( // or /* */) that tend to get generated by templating.
	b = emptyCommentRegex.ReplaceAll(b, []byte{})
	b = emptyBlockCommentRegex.ReplaceAll(b, []byte{})
	// Delete execgen_template build tag.
	b = bytes.ReplaceAll(b, []byte("// +build execgen_template"), []byte{})

	if g.fmtSources {
		oldB := b
		b, err = imports.Process(path, b,
			&imports.Options{Comments: true, TabIndent: true, TabWidth: 2})
		if err != nil {
			// Write out incorrect source for easier debugging.
			b = oldB
			err = errors.Wrap(err, "Code formatting failed with Go parse error")
		}
	}

	// Ignore any write error if another error already occurred.
	_, writeErr := os.Stdout.Write(b)
	if err != nil {
		return err
	}
	return writeErr
}

// usage is a replacement usage function for the flags package.
func (g *execgenTool) usage() {
	fmt.Fprintf(g.stdErr, "Execgen is a tool for generating templated code related to ")
	fmt.Fprintf(g.stdErr, "columnarized execution.\n\n")

	fmt.Fprintf(g.stdErr, "Usage:\n")
	fmt.Fprintf(g.stdErr, "\texecgen [path]...\n\n")

	fmt.Fprintf(g.stdErr, "Supported filenames are:\n")
	for filename := range generators {
		fmt.Fprintf(g.stdErr, "\t%s\n", filename)
	}
	fmt.Fprintf(g.stdErr, "\n")

	fmt.Fprintf(g.stdErr, "Flags:\n")
	g.cmdLine.PrintDefaults()
	fmt.Fprintf(g.stdErr, "\n")
}

func (g *execgenTool) reportError(err error) {
	fmt.Fprintf(g.stdErr, "ERROR: %v\n", err)
}
